# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging
import re

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimDate,
    HprimForeign,
    HprimList,
    HprimMultiple,
    HprimString,
    Segment,
)

logger = logging.getLogger("driverng.drivers.sysmex")


class SysmexSegmentH(HeaderSegment):

    version = HprimString(13, default="A.2")
    date = HprimDate(14, fmt_date="%Y%m%d%H%M")

    patients = HprimForeign("P", "header")
    controls = HprimForeign("S", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class SysmexSegmentP(Segment):

    # Patient ID
    pid = HprimString(3, max_length=16)

    # Alternative Patiet ID
    # Use as Public insurance ID
    apid = HprimString(4, max_length=16)

    prenom = HprimString((6, 1), max_length=20)
    nom = HprimString((6, 2), max_length=20)
    nomjf = HprimString(7, max_length=20)
    date_naissance = HprimDate(8, fmt_date="%Y%m%d")

    # M/F/U
    sexe = HprimString(9, max_length=1)

    # Attending physician ID
    # Primary Doctore Code/Doctor Name
    doctor_code = HprimString((14, 1), max_length=6)
    doctor_name = HprimString((14, 2), max_length=20, trunc=True)

    # Patient height
    # cm (w/ decimals) or feet^inch
    taille = HprimString((17, 1), max_length=6)

    # Patient weight
    # kg (w/ decimals) or pound^ounce
    poids = HprimString((18, 1), max_length=6)

    # In/Out patient differentiation
    # OP: Out patient
    # IP: In patient
    status_admission = HprimString(25, max_length=2)

    # Location
    clinic_code = HprimString((26, 1), max_length=6)
    ward_code = HprimString((26, 2), max_length=6)

    # Date / Time registered
    # Data Registration Date
    # (Patient Data/ Latest Update Date)
    last_updated = HprimDate(33, fmt_date="%Y%m%d")

    orders = HprimForeign("OBR", "patient")


class SysmexSegmentOBR(Segment):

    # Requester specimen ID (HOST side)
    tube_demandeur = HprimString(3, max_length=29)

    # Producer speciment ID (SUIT side)
    tube_executant = HprimString(4, max_length=29)

    tests = HprimList(
        5,
        max_length=200,
        basefield=HprimMultiple(0, code=HprimString(1), nom=HprimString(2)),
    )

    # Processing Priority level:
    # S: STAT/Urgent
    urgence = HprimString(6, max_length=1)

    # Specimen collection or observation date
    cldt = HprimDate(8, fmt_date="%Y%m%d%H%M")

    # Collection volume (Urine etc) Unit ml
    cvlm = HprimString(10, max_length=5)

    # Order Process Code
    # For new samples:
    # A: Add the ordered
    # L: Lab. to obtain a sample
    # For registered samples
    # A: Clear registred orders, register the order as new
    # L: Add orders
    code_action = HprimString(12, max_length=1)

    # Relevent Clinical Information
    rcinf = HprimString(14, max_length=48)

    # Registration Date
    order_date = HprimDate(15, fmt_date="%Y%m%d%H%M")

    # Date and time observation reported or status changed
    dtr = HprimDate(23, fmt_date="%Y%m%d%H%M")

    # Quantity timing / Order Comment Code
    qt = HprimString(28, max_length=6)

    results = HprimForeign("OBX", "order")


class SysmexSegmentOBX(Segment):

    # Test Result Date Format
    # ST: Characters
    # NM: Numeric value
    # CE: Code
    type_resultat = HprimString(3, max_length=2)

    # Test item
    # Code^Name
    test_code = HprimString((4, 1), max_length=30)
    test_name = HprimString((4, 2), max_length=30)

    valeur = HprimString((6, 1), max_length=100)

    # W is set:
    # - When WBC, LYMPH# and LYMPH% are compensated by NRBC
    # - When NEUT# and NEUT% are compensated by IG or NRBC
    # - When WBC is reported by using WBC-D and PLT is reported by using PLT-O
    comment_code = HprimString((6, 2), max_length=3)

    dilution_ratio = HprimString((6, 3), max_length=1)

    # Upper/Lower Flag ~ Delta Check Flag
    # L: Low
    # H: High
    # LL: Panic Low
    # HH: Panic High
    # >: out of linearity
    # W: low reliability
    # A: Abnormal (Except numeric value data)
    abnormal_flags = HprimList(9, max_length=5)

    # P: Preliminary Report
    # F: Final Result
    # I: Pending
    # C: Revision Report
    status = HprimString((12, 1), max_length=1)
    # Edit: input or edit
    # Validate: manual validate
    # Count: count at pad menu
    # Manual Send: Send manually
    # (N/A): Others
    latest_operation = HprimString((12, 2), max_length=20)

    # Sending date
    date = HprimDate(13, fmt_date="%Y%m%d%H%M")

    # Latest operator ID
    operateur = HprimString(17, max_length=6)


class SysmexSegmentQ(Segment):

    tubes = HprimList(4, max_length=200)
    date = HprimDate(7, fmt_date="%Y%m%d%H%M")


class SysmexSegmentS(Segment):

    # Analytical method (Manual or Closed)
    mode = HprimString(3, max_length=10)

    # analyzer Name or Instrument ID
    instrument_id = HprimString((4, 1), max_length=10)
    instrument_name = HprimString((4, 2))

    # Quality Control ("QC" fixed)
    qc = HprimString(7, max_length=2, default="QC")

    # Lot Number of Control Reagent
    ref_lot = HprimString(11, max_length=10)

    # Analysis Item Name
    test = HprimString(12, max_length=8)

    # QC Data
    result = HprimString(13, max_length=8)

    # Analysis Date and Time
    date = HprimDate(16, fmt_date="%Y%m%d%H%M")

    results = HprimForeign("OBX", "control")


class SysmexSegmentC(Segment):

    code1 = HprimString((3, 1))
    code2 = HprimString((3, 2))
    commentaire = HprimString((4, 1), max_length=100)

    def to_string(self, separators=None):
        segment_str = super(SysmexSegmentC, self).to_string(separators=separators)
        if self.commentaire[:3] == "PNG":
            # Cf note plus bas
            # segment_str = segment_str.replace("\\\\", "&R&")
            segment_str = segment_str.replace("/", "&R&")

        return segment_str


class SysmexSegmentL(Segment):

    nombre_segment_patient = HprimString(4)
    nombre_segment = HprimString(5)


class AstmSysmexBase(AstmMessageV2):
    SEGMENT_CLASS_MAP = {
        "H": SysmexSegmentH,
        "P": SysmexSegmentP,
        "OBR": SysmexSegmentOBR,
        "OBX": SysmexSegmentOBX,
        "Q": SysmexSegmentQ,
        "S": SysmexSegmentS,
        "C": SysmexSegmentC,
        "L": SysmexSegmentL,
    }
    DEFAULT_SEPARATORS = "|^~\\&"

    OPTIONS = [
        AstmMessageV2.BooleanOption(
            "legacy_mode_open", label="Legacy mode QC Open", default=True
        ),
        AstmMessageV2.BooleanOption(
            "inverser_nom_prenom", label="Inverser Nom/Prenom", default=False
        ),
        AstmMessageV2.BooleanOption(
            "traiter_comment_obx",
            label="Traiter le commentaire OBX dans le résultat",
            default=False,
        ),
    ]

    def __init__(self, legacy_mode_open=True, inverser_nom_prenom=False, **kwargs):
        self.legacy_mode_open = legacy_mode_open
        self.inverser_nom_prenom = inverser_nom_prenom
        self.instrument_id_seul = False
        self.uf500_tube_inconnu = kwargs.pop("uf500_tube_inconnu", False)
        self.traiter_comment_obx = kwargs.pop("traiter_comment_obx", False)
        super(AstmSysmexBase, self).__init__(**kwargs)

    def to_string(self, message, *args, **kwargs):
        message.end.nombre_segment_patient = len(message.patients)
        message.end.nombre_segment = len(message.get_segments())
        return super(AstmSysmexBase, self).to_string(message, *args, **kwargs)

    def parse_segment(self, line, separators):
        # \ et & ignorés
        line = line.replace("&R&", "/")
        line = line.replace(separators[3], "{0}E{0}".format(separators[3]))
        line = line.replace(separators[4], "{0}T{0}".format(separators[3]))
        return super(AstmSysmexBase, self).parse_segment(line, separators)

    def create(self, **kwargs):
        message = self.SEGMENT_CLASS_MAP["H"](type_segment="H", **kwargs)
        message.end = self.SEGMENT_CLASS_MAP["L"](type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        message = self.create(date=context["now"])
        actions = {worklist.ELSE: "L"}
        if self.HANDLE_DELETED:
            actions["del"] = "C"
        else:
            actions["del"] = worklist.IGNORE
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action=actions,
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
        ):
            physician = [p for p in patient.physicians if p.type.lower() == "attending"]

            if physician and (physician[0].last_name, physician[0].first_name) != (
                None,
                None,
            ):
                doctor_code = physician[0].code
                doctor_name = "{} {}".format(
                    physician[0].last_name or "", physician[0].first_name or ""
                )
            else:
                doctor_code = patient.physician or ""
                if patient.physician_name:
                    doctor_name = "{} {}".format(
                        patient.physician, patient.physician_name
                    )
                else:
                    doctor_name = ""
            seg_p = message.patients.add(
                pid=patient.patient_id,
                apid=patient.patient_id,
                nomjf=patient.maiden_name,
                date_naissance=patient.birthdate,
                sexe=patient.sex,
                doctor_code="{:.6s}".format(doctor_code),
                doctor_name=doctor_name,
                clinic_code=patient.services[0].code if patient.services else "",
            )
            if self.inverser_nom_prenom:
                seg_p.nom = patient.first_name
                seg_p.prenom = patient.last_name
            else:
                seg_p.nom = patient.last_name
                seg_p.prenom = patient.first_name
            for sample, action, priority in patient.samples:
                seg_o = seg_p.orders.add(
                    tube_demandeur=sample.sample_id,
                    urgence=priority,
                    code_action=action,
                    rcinf=sample.patient_comment1,
                    # SUIT+ only
                    doctor_code="{:.6s}".format(doctor_code),
                    doctor_name=doctor_name,
                    clinic_code=patient.services[0].code if patient.services else "",
                    ward_code=patient.services[0].code if patient.services else "",
                    ward_name=patient.services[0].name if patient.services else "",
                    cldt=sample.collection_date,
                )
                # C'est pour gérer un hack du paramétrage,
                # plusieurs code_demande dans le code_demande.
                for order in sample.orders:
                    test = seg_o.tests.add()
                    test.code = order.code

        if message.patients:
            return message
        elif context["query"]:
            # envoyer par défaut les codes demandesi RBC~XTAL~EC~BACT~WBC
            # lorsque le tube est inconnu (à tester)

            for seg_q in context["query"].queries:
                for sample_id in seg_q.tubes:
                    seg_p = message.patients.add()
                    seg_o = seg_p.orders.add(tube_demandeur=sample_id)
                    if self.uf500_tube_inconnu:
                        for code in ["RBC", "XTAL", "EC", "BACT", "WBC"]:
                            test = seg_o.tests.add()
                            test.code = code
            return message

    def get_tube_graphe(self, commentaire, tube):
        """Extrait le tube et le graphe d'un champ commentaire
        @param commentaire: le champ à parser
        @param tube: le tube_executant (numero d'echantillon)
        @return (tube, graph_image_name)
        """
        match = re.match(
            r"^.*(?# des donnees dont on n'a pas besoin)"
            "\\d{4}_(\\d{2}_){4}(?# une date au format YYYY_MM_DD_hh_mm"
            "cette date permet de se reperer, les donnees importantes"
            "sont juste apres)"
            "(?P<tube>[^_]+)_(?# le numero d'echantillon, delimite par des `_`)"
            "(?P<graph_image_name>.*)$(?# enfin, le nom de l'image du graphe.)",
            commentaire,
        )
        if match:
            return match.group("tube"), match.group("graph_image_name")
        logger.warning(
            "Impossible d'extraire un tube connu dans '{}'".format(commentaire)
        )
        return (None, None)

    @staticmethod
    def get_lots_reactifs(seg_obx):
        raise NotImplementedError

    @staticmethod
    def ignore_segment_obx(seg_obx):
        raise NotImplementedError

    @staticmethod
    def get_flags(message):
        return {}

    def get_analyseur_rack(self, seg_obr):
        raise NotImplementedError

    def get_analyseur(self, analyseur, seg_obx):
        raise NotImplementedError

    def get_analyseur_qc(self, message):
        raise NotImplementedError

    def get_alarmes(self, seg_obx):
        raise NotImplementedError

    def get_comment_obx(self, seg_obx, alarmes=[]):
        raise NotImplementedError

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_q in message.queries:
                for tube in seg_q.tubes:
                    worklist.get_or_create(type="patient", sample_id=tube)
            return worklist

        worklist = Worklist(context="result")
        self._decode_patients(worklist, message)
        self._decode_QC(worklist, message)
        return worklist

    def _decode_patients(self, worklist, message):
        flags = self.get_flags(message)
        for seg_p in message.patients:
            for seg_obr in seg_p.orders:

                # if there's ne result associated to this frame, we mustn't
                # create the sample.
                # It will be created if we find results associated with this
                # frame in the further orb frames.
                if not seg_obr.results:
                    continue

                analyseur, rack = self.get_analyseur_rack(seg_obr)
                sample = worklist.get_or_create(
                    type="patient", sample_id=seg_obr.tube_executant, rack=rack
                )
                for seg_obx in seg_obr.results:
                    if self.ignore_segment_obx(seg_obx):
                        continue
                    alarmes = self.get_alarmes(seg_obx)
                    alarmes.extend(
                        flags.get(seg_obr.tube_executant, {}).get(seg_obx.test_code, [])
                    )
                    result_context = {
                        key: value
                        for key, value in (
                            ("lots", self.get_lots_reactifs(seg_obx, "reagent")),
                            ("instrument", self.get_analyseur(analyseur, seg_obx)),
                        )
                        if value is not None
                    }
                    if self.traiter_comment_obx:
                        comments = self.get_comment_obx(seg_obx, alarmes)
                        if comments:
                            result_context["comments"] = comments

                    sample.results.add(
                        code=seg_obx.test_code,
                        value=seg_obx.valeur,
                        date=seg_obx.date,
                        flags=alarmes,
                        **result_context
                    )

                # Gestion des commentaires sur tube (alarmes) et graphes
                for seg_obr in seg_p.orders:
                    commentaire_tube = []
                    for seg_obx in seg_obr.results:
                        instrument = self.get_analyseur(analyseur, seg_obx)
                        for seg_c in seg_obx.commentaires:
                            if seg_c.commentaire.startswith("PNG"):
                                tube, code_test = self.get_tube_graphe(
                                    seg_c.commentaire, seg_obr.tube_executant
                                )
                                if tube is not None and code_test is not None:
                                    r_kw = {
                                        "code": code_test,
                                        "value": seg_c.commentaire,
                                    }
                                    if instrument:
                                        r_kw["instrument"] = instrument
                                    worklist.get_or_create(
                                        type="patient", sample_id=tube, rack=rack
                                    ).results.add(**r_kw)
                            elif seg_c.commentaire:  # commentaire tube
                                commentaire_tube.append(seg_c.commentaire)

                    if commentaire_tube:
                        sample.results.add(
                            code="ALARME", value="", comments=commentaire_tube
                        )

    def _decode_QC(self, worklist, message):
        if not message.controls:
            return

        # dans le precedent code, apres la boucle `for` principale, on prennait
        # le dernier analyseur, et l'attribuait a chaque resultat sans
        # analyseur.
        # Maintenant, nous prenons le premier analyseur, en partant de la fin
        # et l'attribuons dans la boucle principale aux résultats qui
        # n'ont pas d'analyseur. Cela réduit le nombre d'ittérations, car
        # l'analyseur est toujours sur le dernier resultat.
        for seg_s in reversed(message.controls):
            default_analyseur = self.get_analyseur_qc(message, seg_s)
            if default_analyseur:
                break

        for seg_s in message.controls:

            analyseur = self.get_analyseur_qc(message, seg_s) or default_analyseur
            # On virre le prefixe QC-
            if seg_s.ref_lot[:3] == "QC-":
                ref_lot = seg_s.ref_lot[3:]
            else:
                ref_lot = seg_s.ref_lot

            if seg_s.test in (
                "H_RACK",
                "U_RACK",
                "C_RACK",
                "H_TUBE",
                "U_TUBE",
                "C_TUBE",
                "H_INID",
                "U_INID",
                "C_INID",
                "H_INST",
                "U_INST",
                "C_INST",
            ):
                # On ignore ce segment
                continue

            if self.legacy_mode_open:
                if not analyseur and seg_s.mode == "Manual":
                    instrument = "{}-OPEN".format(seg_s.instrument_id)
                else:
                    instrument = analyseur or seg_s.instrument_id
            else:
                instrument = analyseur or seg_s.instrument_id
                if seg_s.mode in ("Manual", "Open"):
                    instrument += "-OPEN"

            if len(ref_lot) == 8:
                # nom_lot + control level
                nom_lot = ref_lot[:4]
                control_level = ref_lot[4:]
            else:
                nom_lot = control_level = None

            # elimination of empty attributes
            sample_attributes = {
                key: value
                for key, value in (
                    ("control_level", control_level),
                    ("control_name", nom_lot),
                    ("control_lot", ref_lot),
                )
                if value is not None
            }
            # there is no sample ID because, there's not OBR frame
            # there's only H, L and C frames.
            worklist.get_or_create(
                type="control", sample_id="", **sample_attributes
            ).results.add(
                code=seg_s.test,
                value=seg_s.result,
                date=seg_s.date,
                instrument=instrument,
            )
